
/*
<Arduino>        <GY-85>
   A5              SCL
   A4              SDA
   3.3V            3.3V (ou VCC)
   GND             GND

Note : les pins A4=SDA et A5=SCL sont obligatoires pour utiliser les librairies proposées ici.
       Si vous devez utiliser ces pins (A4 et/ou A5) pour d'autres modules alors il faudra aller dans le code des librairies (celles-ci ou les autres, au choix) et changer les pins cibles qui sont hardcodés.
*/


// librairies
#include <Wire.h>     // librairie Wire     : accès registre composants (général)
#include <ADXL345.h>  // librairie ADXL345  : accéléromètre (spécifique)
#include <ITG3200.h>  // librairie ITG3200  : gyroscope (spécifique)
#include <HMC5883L.h> // librairie HMC5883L : magnétomètre (spécifique)



// VARIABLES //

// constantes générales
const int DELAY_FIN_LOOP_MS = 10; // 100 / 20 / 0 // attente forcée à la fin de chaque appel à la fonction loop(), en millisecondes

// communication avec Processing
const boolean IS_DATA_TO_SEND_TO_PROCESSING = true; // true = Processing = humainement incompréhensible, optimisé, une suite de nombres avec séparateurs sur une seule ligne qui n'arrête jamais // false = présentation moniteur série compréhensible, avec sauts de ligne
const char SEPARATEUR_VALEURS_ENVOYEES = ';'; // caractère séparateur de variables (utilisé n fois par groupe de données)
const char SEPARATEUR_FINAL_VALEURS_ENVOYEES = '/'; // caractère séparateur de fin de groupe (utilisé 1 seule fois par groupe de données)

// constantes de mesure
const int NBR_MESURES_CALIBRATION = 2000; // 200 // nombre de mesures par axe afin de bien calibrer les composants (leur valeur par défaut, car les résultats seront ensuite calculés en fonction de ces valeurs considérées comme neutres)

// variables de librairies
ADXL345 acc; // accéléromètre
ITG3200 gyro = ITG3200(); // gyroscope
HMC5883L compass; // magnétomètre

// variables déclarées dans le tuto pour reprise bouts de code
float  gx, gy, gz;
float  gx_rate, gy_rate, gz_rate;
int ix, iy, iz;
float anglegx = 0.0, anglegy = 0.0, anglegz = 0.0;
int ax, ay, az;
//int rawX, rawY, rawZ;
float rawX, rawY, rawZ;
float X, Y, Z;
float rollrad, pitchrad;
float rolldeg, pitchdeg;
int error = 0;
float aoffsetX, aoffsetY, aoffsetZ;
float goffsetX, goffsetY, goffsetZ;
unsigned long time, looptime;

// variables supplémentaires
float rawZcalc;
float vitesseX = 0, vitesseY = 0, vitesseZ = 0;
float lastVitesseX = 0, lastVitesseY = 0, lastVitesseZ = 0;
float positionX = 0, positionY = 0, positionZ = 0;
float lastPositionX = 0, lastPositionY = 0, lastPositionZ = 0;
float lastRawX, lastRawY, lastRawZ;
//float fa[3][2], fv[3][3], fp[3][4]; // fonctions accélération, vitesse et position (save uniquement du coef de chaque degré)
unsigned long timerMesuresAccelerometre = 0, lastTimerMesuresAccelerometre = 0;
float coef;



// fonction setup() : configuration, appelée 1 fois à l'initialisation
void setup(){
  // déclaration de communication série
  Serial.begin(9600);

  // activation de l'accéléromètre
  acc.powerOn();

  // activation du magnétomètre
  compass = HMC5883L();
  error = compass.SetScale(1.3); // Set the scale to +/- 1.3 Ga of the compass // cf. datasheet pour adapter la finesse de perception du capteur en fonction du besoin // attention : cf. datasheet pour avoir la liste exhaustive des valeurs autorisées !
  // écriture de l'erreur s'il y en a une
  if(error != 0){
    Serial.println(compass.GetErrorText(error));
  }
  error = compass.SetMeasurementMode(Measurement_Continuous); // Set the measurement mode to Continuous
  if(error != 0){
    Serial.println(compass.GetErrorText(error));
  }
  
  // calibration des capteurs (accéléromètre + gyroscope)
  for (int i = 0; i < NBR_MESURES_CALIBRATION; i++) {
    acc.readAccel(&ax, &ay, &az); // prise de mesure accéléromètre
    gyro.readGyro(&gx, &gy, &gz); // prise de mesure gyroscope
    if(i > 0){
      // moyenne entre la valeur en cours et la précédente (qui était elle-même une moyenne de deux valeurs dont une moyenne, etc) //
      // accéléromètre
      aoffsetX = (ax + aoffsetX) / 2;
      aoffsetY = (ay + aoffsetY) / 2;
      aoffsetZ = (az + aoffsetZ) / 2;
      // gyroscope
      goffsetX = (gx + goffsetX) / 2;
      goffsetY = (gy + goffsetY) / 2;
      goffsetZ = (gz + goffsetZ) / 2;
    }else{
      // initialisation //
      // accéléromètre
      aoffsetX = ax;
      aoffsetY = ay;
      aoffsetZ = az;
      // gyroscope
      goffsetX = gx;
      goffsetY = gy;
      goffsetZ = gz;
    }
  }
  
  // initialisation du gyroscope
  delay(1000);
  gyro.init(ITG3200_ADDR_AD0_LOW);
  //gyro.zeroCalibrate(2500, 2);
} // fin setup()



// fonction loop() : fonction principale, appelée en boucle à l'infini
void loop(){
  
  // lecture des données du GY-85 (via librairies) //
  
  // accéléromètre //
  
  // lecture données accéléromètre
  acc.readAccel(&ax, &ay, &az);

  // calcul de la différence entre la donnée réelle et la donnée neutre
  rawX = ax - aoffsetX;
  rawY = ay - aoffsetY;
  rawZcalc = az  - (255 - aoffsetZ);
  rawZ = az  - aoffsetZ;
  
  // variables intermédiaires pour le calcul des angles
  X = rawX/256.00;
  Y = rawY/256.00;
  //Z = rawZ/256.00;
  Z = rawZcalc/256.00;
  
  // calcul des angles depuis les données de l'accéléromètre
  //rollrad = atan(Y/sqrt(X*X+Z*Z)); // roll (radians)
  //pitchrad = atan(X/sqrt(Y*Y+Z*Z)); // pitch (radians)
  rolldeg = 180*(atan(Y/sqrt(X*X+Z*Z)))/PI; // roll (degrés)
  pitchdeg = 180*(atan(X/sqrt(Y*Y+Z*Z)))/PI; // pitch (degrés)
  
  // code tuto (pour info, code tel que je l'ai trouvé à la base)
  //rawX = rawX * 0.0039;
  //rawY = rawY * 0.0039;
  //rawZ = rawZ * 0.0039;
  //rolldeg = atan2(rawY, rawZ) * 180/PI;
  //pitchdeg = atan2(-rawX, sqrt(rawY*rawY + rawZ*rawZ)) * 180/PI;
  
  
  // gyroscope //
  
  // date actuelle
  time = millis();
  
  // lecture données gyroscope
  gyro.readGyro(&gx, &gy, &gz);

  // calcul du temps mis pour la prise de mesure du gyroscope
  looptime = millis() - time;

  // calcul de (??? angle/force ???) par rapport à la fréquence de capture des données du gyroscope (???) [à vérifier pour avoir la bonne explication]
  //gx_rate = (gx-goffsetX) / 14.375;
  //gy_rate = (gy-goffsetY) / 14.375;
  //gz_rate = (gz-goffsetZ) / 14.375;
  // application de coefs persos car les fréquences données par le manuel ne sont pas correctes pour mon gyroscope, un décalage important arrivait sans arrêt (de façon constante/linéaire, c'est déjà ça)
  gx_rate = (gx-goffsetX) / 14.375 - 0.1275;
  gy_rate = (gy-goffsetY) / 14.375 - 0.0625;
  gz_rate = (gz-goffsetZ) / 14.375 - 0.0145;
  
  // mise à jour des angles (gyroscope) par rapport au temps qu'a pris la mesure
  anglegx += gx_rate * looptime;
  //anglegy += gx_rate * looptime;
  anglegy += gy_rate * looptime;
  anglegz += gz_rate * looptime;
  
  // température du gyroscope (?!?) [pour info, non utilisé pour l'instant mais disponible !]
  // Temperature in °C = 35 + (Measurement Value + 13200) / 280;
  
  
  // magnétomètre //

  // note : apparemment on peut lire et exploiter n'importe laquelle des 3 données disponibles (liées aux 3 angles). Ici on prendra le raw, l'angle autour de l'axe X.
  // Attention : si le futur robot évolue sur un plan parallèle aux vecteurs X et Y du capteur, alors il ne faut pas prendre le Z qui sera toujours orthogonal à la direction du champ magnétique terrestre et qui ne captera rien.
  //             si le capteur n'est pas orienté selon la norme XYZ standard (X et Y parallèles au plan d'évolution), il faudra peut-être prendre un autre axe que X comme repère, en fonction du cas.
  //             si le futur robot évolue dans l'espace (pouvant se tourner sur lui-même sur 3 axes), il faudra prendre les 3 axes en compte, analyser les résultats et trouver un algorythme qui s'adapte.
  
  // lecture donnée magnétomètre
  MagnetometerRaw raw = compass.ReadRawAxis();

  // mise à l'échelle (?)
  MagnetometerScaled scaled = compass.ReadScaledAxis();

  // calcul des forces (via librairie)
  int MilliGauss_OnThe_XAxis = scaled.XAxis;

  // calcul de l'angle du capteur
  float heading = atan2(scaled.YAxis, scaled.XAxis);
  float declinationAngle = 0.0457;
  // angle final magnétomètre (en radians, attention au modulo 2*PI)
  heading += declinationAngle;
  if(heading < 0)
    heading += 2*PI;
  if(heading > 2*PI)
    heading -= 2*PI;
  
  // angle final magnétomètre (degrés)
  float headingDegrees = heading * 180 / M_PI;
  
  
  
  // envoi des données à Processing ou output des variables //

  // note : j'ai appliqué quelques coefficients qui sont bien trouvés mais pas très scientifiques/mathématiques. On en parlera si vous voulez. Le truc c'est que les capteurs ne sont pas très précis et il faut parfois utiliser la ruse...
  //        Il faut encore trouver la fonction mathématique qui ira bien pour le magnétomètre, c'est une fonction de degré 2 et je ne m'y suis pas encore penché mais ça devrait être simple, on pourra aussi en parler.
  
  // accéléromètre
  if(IS_DATA_TO_SEND_TO_PROCESSING){
    Serial.print(rawX);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    Serial.print(rawY);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    Serial.print(rawZ);
    //Serial.print(rawZ * 0.0039);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    //Serial.print(rolldeg * 12.7);
    Serial.print(rolldeg * 3.5);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    //Serial.print(pitchdeg * 12.7);
    Serial.print(pitchdeg * 3.5);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
  }else{
    Serial.print(rawX);
    Serial.print("\t");
    Serial.print(rawY);
    Serial.print("\t");
    Serial.print(rawZ);
    Serial.print("\t");
    Serial.print(rolldeg);
    Serial.print("\t");
    Serial.print(pitchdeg);
    Serial.print("\t");
  }
  
  // gyroscope
  if(IS_DATA_TO_SEND_TO_PROCESSING){
    //Serial.print(anglegx);
    Serial.print(-anglegx);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    Serial.print(anglegy);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
    //Serial.print(anglegz);
    Serial.print(-anglegz);
    Serial.print(SEPARATEUR_VALEURS_ENVOYEES);
  }else{
    Serial.print(anglegx);
    Serial.print("\t");
    Serial.print(anglegy);
    Serial.print("\t");
    Serial.print(anglegz);
    Serial.print("\t");
  }
  
  // magnétomètre
  if(IS_DATA_TO_SEND_TO_PROCESSING){
    Serial.print(headingDegrees);
    Serial.print(SEPARATEUR_FINAL_VALEURS_ENVOYEES);
  }else{
    Serial.print(headingDegrees);
    Serial.println();
  }
  
  
  
  // attente de fin de loop()
  delay(DELAY_FIN_LOOP_MS);
}







